@@ -21,7 +21,7 @@
 
 /*
  * connection.c -- defines the connection object DBAPI 2.0
- * $Id: connection.c,v 1.47.2.5 2002/10/25 10:12:56 fog Exp $
+ * $Id: connection.c,v 1.51 2003/03/06 10:13:15 fog Exp $
  */
 
 #include <assert.h>
@@ -71,7 +71,8 @@ curs_commitall(connobject *self)
     for (i = 0; i < len; i++) {
         cursor = (cursobject *)PyList_GetItem(cursors, i);
         assert(cursor);
-        if (cursor->keeper->status == KEEPER_BEGIN && cursor->autocommit == 0){
+        if (cursor->keeper->status == KEEPER_BEGIN
+            && cursor->isolation_level > 0){
             pthread_mutex_lock(&(cursor->keeper->lock));
             if (cursor->keeper->status == KEEPER_BEGIN) {
                 cursor->keeper->status = KEEPER_CONN_LOCK;
@@ -132,7 +133,8 @@ curs_rollbackall(connobject *self)
     for (i = 0; i < len; i++) {
         cursor = (cursobject *)PyList_GetItem(cursors, i);
         assert(cursor);
-        if (cursor->keeper->status == KEEPER_BEGIN && cursor->autocommit == 0){
+        if (cursor->keeper->status == KEEPER_BEGIN
+            && cursor->isolation_level > 0){
             pthread_mutex_lock(&(cursor->keeper->lock));
             if (cursor->keeper->status == KEEPER_BEGIN) {
                 cursor->keeper->status = KEEPER_CONN_LOCK;
@@ -185,8 +187,7 @@ _psyco_conn_close(connobject *self)
     PyObject *tmpobj = NULL;
     connkeeper *keeper;
     
-    Dprintf("_psyco_conn_close(): shutting down postgres connections\n");
-
+    Dprintf("_psyco_conn_close(): closing all cursors\n");
     curs_closeall(self);
 
     /* orphans all the children cursors but do NOT destroy them (note that
@@ -199,11 +200,14 @@ _psyco_conn_close(connobject *self)
     for (i = len-1; i >= 0; i--) {
         tmpobj = PyList_GetItem(self->cursors, i);
         assert(tmpobj);
+        Dprintf("_psyco_conn_close(): detaching cursor at %p: refcnt = %d\n",
+                tmpobj, tmpobj->ob_refcnt);
         Py_INCREF(tmpobj);
         PySequence_DelItem(self->cursors, i);
+        dispose_pgconn((cursobject *)tmpobj);
         ((cursobject *)tmpobj)->conn = NULL; /* orphaned */
-        Dprintf("_psyco_conn_close(): cursor refcnt = %d\n",
-                tmpobj->ob_refcnt);
+        Dprintf("_psyco_conn_close(): cursor at %p detached: refcnt = %d\n",
+                tmpobj, tmpobj->ob_refcnt);
     }
 
     /* close all the open postgresql connections */
@@ -231,6 +235,7 @@ _psyco_conn_close(connobject *self)
 
     /* orphan default cursor and destroy it (closing the last connection to the
        database) */
+    Dprintf("_psyco_conn_close(): killing stdmanager\n");
     self->stdmanager->conn = NULL;
     Py_DECREF(self->stdmanager);
 }
@@ -324,6 +329,45 @@ psyco_conn_cursor(connobject *self, PyObject *args)
 }
 
 
+/* psyco_conn_set_isolation_level() - set isolation level for this connection
+   and all derived cursors */
+
+static char psyco_conn_set_isolation_level__doc__[] = 
+"Switch connection (and derived cursors) isolation level.";
+
+static void
+_psyco_conn_set_isolation_level(connobject *self, int level)
+{
+    PyObject *tmpobj;
+    int i, len;
+    
+   /* save the value */
+    self->isolation_level = (int)level;
+
+    /* call set_isolation_level on all the cursors to change their status */
+    len = PyList_Size(self->cursors);
+    for (i = 0; i < len; i++) {
+        tmpobj = PyList_GetItem(self->cursors, i);
+        assert(tmpobj);
+        curs_switch_isolation_level((cursobject *)tmpobj, level);
+    }
+}
+
+static PyObject *
+psyco_conn_set_isolation_level(connobject *self, PyObject *args)
+{
+    long int level;
+    
+    if (!PyArg_ParseTuple(args, "l", &level)) {
+        return NULL;
+    }
+
+    _psyco_conn_set_isolation_level(self, level);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 /* psyco_conn_autocommit() - put connection (and all derived cursors)
    in autocommit mode */
 
@@ -333,24 +377,15 @@ static char psyco_conn_autocommit__doc__[] =
 static PyObject *
 psyco_conn_autocommit(connobject *self, PyObject *args)
 {
-    PyObject *tmpobj;
     long int ac = 1; /* the default is to set autocommit on */
-    int i, len;
+    int isolation_level = 0;
     
     if (!PyArg_ParseTuple(args, "|l", &ac)) {
         return NULL;
     }
 
-    /* save the value */
-    self->autocommit = (int)ac;
-
-    /* call autocommit(ac) on all the cursors to change their status too */
-    len = PyList_Size(self->cursors);
-    for (i = 0; i < len; i++) {
-        tmpobj = PyList_GetItem(self->cursors, i);
-        assert(tmpobj);
-        curs_switch_autocommit((cursobject *)tmpobj, ac);
-    }
+    if (ac == 0) isolation_level = 2;
+    _psyco_conn_set_isolation_level(self, isolation_level);
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -393,6 +428,8 @@ static struct PyMethodDef psyco_conn_methods[] = {
      METH_VARARGS, psyco_conn_cursor__doc__},
     {"autocommit", (PyCFunction)psyco_conn_autocommit,
      METH_VARARGS, psyco_conn_autocommit__doc__},
+    {"set_isolation_level", (PyCFunction)psyco_conn_set_isolation_level,
+     METH_VARARGS, psyco_conn_set_isolation_level__doc__},   
     {"serialize", (PyCFunction)psyco_conn_serialize,
      METH_VARARGS, psyco_conn_serialize__doc__},
     {NULL, NULL}
@@ -448,7 +485,7 @@ psyco_conn_destroy(connobject *self)
 static char Conntype__doc__[] = "Connections Object.";
 
 static PyTypeObject Conntype = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -487,7 +524,7 @@ new_psyco_connobject(char *dsn, int maxconn, int minconn, int serialize)
 
     Dprintf("new_psyco_connobject(): creating new connection\n");
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     Conntype.ob_type = &PyType_Type;
 #endif
@@ -502,7 +539,7 @@ new_psyco_connobject(char *dsn, int maxconn, int minconn, int serialize)
     self->cursors = PyList_New(0);
     self->avail_conn = PyList_New(0);
     self->closed = 0;
-    self->autocommit = 0;
+    self->isolation_level = 2;
     self->serialize = serialize;
     
     /* allocate default manager thread and keeper */
