@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2001 Michele Comitini <mcm@initd.net>
- * Copyright (C) 2001-2002 Federico Di Gregorio <fog@debian.org>
+ * Copyright (C) 2001-2003 Federico Di Gregorio <fog@debian.org>
  *
  * This file is part of the psycopg module.
  *
@@ -21,7 +21,7 @@
 
 /*
  * cursor.c - defines the cursor object DBAPI 2.0
- * $Id: cursor.c,v 1.102.2.12 2002/10/25 10:12:56 fog Exp $
+ * $Id: cursor.c,v 1.109 2003/05/07 13:27:17 fog Exp $
  */
 
 #include <assert.h>
@@ -41,6 +41,9 @@
 #include <postgres.h>
 #endif
 
+/* include pgtypes.h to get right types #defines */
+#include "pgtypes.h"
+
 
 /**** UTILITY FUNCTIONS (callable from C only) ****/
 
@@ -183,7 +186,6 @@ _psyco_curs_destroy(cursobject *self)
     /* remove the cursor from the connection list */
     if (self->conn) {
         pthread_mutex_lock(&(self->conn->lock));
-        /* Py_INCREF(self); */
         _psyco_curs_getout(self);
         pthread_mutex_unlock(&(self->conn->lock));
     }
@@ -196,6 +198,9 @@ _psyco_curs_destroy(cursobject *self)
     Py_XDECREF(self->description);
     Py_INCREF(Py_None);
     self->description = Py_None;
+    Py_XDECREF(self->status);
+    Py_INCREF(Py_None);
+    self->status = Py_None;
     Py_XDECREF(self->casts);
     self->casts = NULL;
     if (self->notice) free(self->notice);
@@ -260,10 +265,9 @@ commit_pgconn(cursobject *self)
     PGresult *pgres = NULL;
     
     /* if in autocommit mode does nothing */
-    Dprintf("commit_pgconn(): pgconn = %p, autocommit = %d, status = %d\n",
-            self->pgconn, self->autocommit, self->keeper->status);
-    if (self->autocommit == 1
-        || self->keeper->status != KEEPER_BEGIN)
+    Dprintf("commit_pgconn(): pgconn = %p, level = %d, status = %d\n",
+            self->pgconn, self->isolation_level, self->keeper->status);
+    if (self->isolation_level == 0 || self->keeper->status != KEEPER_BEGIN)
         return 0;
     
     assert(self->pgconn);
@@ -302,20 +306,25 @@ commit_pgconn(cursobject *self)
 int
 begin_pgconn(cursobject *self)
 {
-    const char *query = "BEGIN; SET TRANSACTION ISOLATION LEVEL SERIALIZABLE";
+    const char *query[] = {
+        NULL,
+        "BEGIN; SET TRANSACTION ISOLATION LEVEL READ COMMITTED",
+        "BEGIN; SET TRANSACTION ISOLATION LEVEL SERIALIZABLE",
+        "BEGIN; SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"
+    };
+
     int pgstatus, retvalue = -1;
     PGresult *pgres = NULL;
 
     /* if in autocommit mode does nothing */
-    Dprintf("begin_pgconn(): pgconn = %p, autocommit = %d, status = %d\n",
-            self->pgconn, self->autocommit, self->keeper->status);
-    if (self->autocommit == 1
-        || self->keeper->status != KEEPER_READY)
+    Dprintf("begin_pgconn(): pgconn = %p, level = %d, status = %d\n",
+            self->pgconn, self->isolation_level, self->keeper->status);
+    if (self->isolation_level == 0 || self->keeper->status != KEEPER_READY)
         return 0;
     
     assert(self->pgconn);
 
-    pgres = PQexec(self->pgconn, query);
+    pgres = PQexec(self->pgconn, query[self->isolation_level]);
     if (pgres == NULL) {
         pgconn_set_critical(self);
         goto cleanup;
@@ -354,20 +363,12 @@ abort_pgconn(cursobject *self)
     PGresult *pgres = NULL;
 
     /* if in autocommit mode does nothing */
-    Dprintf("abort_pgconn(): pgconn = %p, autocommit = %d, status = %d\n",
-            self->pgconn, self->autocommit, self->keeper->status);
-    if (self->autocommit == 1
-        || self->keeper->status != KEEPER_BEGIN)
+    Dprintf("abort_pgconn(): pgconn = %p, level = %d, status = %d\n",
+            self->pgconn, self->isolation_level, self->keeper->status);
+    if (self->isolation_level == 0 || self->keeper->status != KEEPER_BEGIN)
         return 0;
 
     assert(self->pgconn);
-    /* if (PQstatus(self->pgconn) != CONNECTION_OK){
-        char *err = PQerrorMessage(self->pgconn);
-        Dprintf("abort_pgconn(): connection NOT OK: %s\n", err);
-        PyErr_SetString(OperationalError, err);
-        goto cleanup;
-    }
-    Dprintf("abort_pgconn(): connection at %p OK\n", self->pgconn); */
     
     pgres = PQexec(self->pgconn, query);
     if (pgres == NULL) {
@@ -412,8 +413,11 @@ alloc_keeper(connobject *conn)
     
     Dprintf("alloc_keeper(): opening new postgresql connection\n");
     Dprintf("alloc_keeper(): dsn = %s (%p)\n", conn->dsn, conn->dsn);
-    
+
+    Py_BEGIN_ALLOW_THREADS;
     pgconn = PQconnectdb(conn->dsn);
+    Py_END_ALLOW_THREADS;
+        
     Dprintf("alloc_keeper(): new posgresql connection at %p\n", pgconn);
         
     if (pgconn == NULL)
@@ -575,31 +579,29 @@ request_pgconn(cursobject *self)
 
 
 
-/* curs_switch_autocommit() - switch from and back to autocommit
+/* curs_switch_isolation level() - switch isolation level
  *
- * TODO: develop a clean autocommit policy
+ * TODO: develop a clean switching policy
  */
 void
-curs_switch_autocommit(cursobject *self, long int ac)
+curs_switch_isolation_level(cursobject *self, long int level)
 {
     pthread_mutex_lock(&(self->keeper->lock));
     Py_BEGIN_ALLOW_THREADS;
     
     /* if the cursor was not in autocommit and we set autocommit we need to
        ABORT the transaction */
-    if (self->autocommit == 0 && ac != 0) {
-        if (abort_pgconn(self) < 0) goto error;
-        self->autocommit = 1;
-    }
-
-    /* else we simply set autocommit to 0 */
-    if (self->autocommit != 0 && ac == 0) {
-        self->autocommit = 0;
+    if (self->isolation_level > 0 && level == 0) {
+        if (abort_pgconn(self) < 0) goto error;    
     }
 
+    /* else we simply set the right isolation level (the first call to
+       execute will begin the right transaction */
+    self->isolation_level = level;
+    
  error:
-    Dprintf("curs_switch_autocommit(): pgconn = %p, autocommit = %d, "
-            "status = %d\n", self->pgconn, self->autocommit,
+    Dprintf("curs_switch_isolation_level(): pgconn = %p, level = %d, "
+            "status = %d\n", self->pgconn, self->isolation_level,
             self->keeper->status);
     pthread_mutex_unlock(&(self->keeper->lock));
     Py_END_ALLOW_THREADS;
@@ -625,6 +627,10 @@ psyco_curs_reset(cursobject *self, int resetconn)
     Py_INCREF(Py_None);
     self->description = Py_None;
 
+    Py_XDECREF(self->status);
+    Py_INCREF(Py_None);
+    self->status = Py_None;
+
     Py_XDECREF(self->casts);
     self->casts = NULL;
 
@@ -706,9 +712,11 @@ psyco_curs_commit(cursobject *self, PyObject *args) {
 /* _psyco_curs_execute() - execute a query and parse results, used by both the
    .execute() and the .callproc() methods */
 static PyObject *
-_psyco_curs_execute(cursobject *self, char *query)
+_psyco_curs_execute(cursobject *self, char *query,
+                    _psyco_curs_execute_callback cb, PyObject *cb_args)
 {
     int pgstatus, old_keeper_status;
+    PyObject *res = NULL;
     
     /* even if we fail, we remove any information about the previous query */
     psyco_curs_reset(self, 0);
@@ -750,22 +758,38 @@ _psyco_curs_execute(cursobject *self, char *query)
     }
 
     pgstatus = PQresultStatus(self->pgres);
-
+    
     switch(pgstatus) {
 
         /* send data to the backend */
     case PGRES_COPY_OUT:
-        break;
+        Dprintf("_psyco_curs_execute(): command returned COPY_OUT\n");
+        /* fall through to the COPY_IN */
 
         /* data from the backend */
     case PGRES_COPY_IN:
+        Dprintf("_psyco_curs_execute(): command returned COPY_IN\n");
+        self->rowcount = 0;
+        Py_XDECREF(self->description);
+        Py_INCREF(Py_None);
+        self->description = Py_None;
+        Py_XDECREF(self->status);
+        Py_INCREF(Py_None);
+        self->status = Py_None;
+        Py_XDECREF(self->casts);
+        self->casts = NULL;
+        self->notuples = 1;
+
+        /* call the callback */
+        res = cb(self, cb_args);
         break;
 
         /* tuples, this was a select */
     case PGRES_TUPLES_OK: {
         int i, pgnfields = PQnfields(self->pgres);
         int pgbintuples = PQbinaryTuples(self->pgres);
-        
+        int *dsize = NULL;
+
         self->notuples = 0;
         self->rowcount = PQntuples(self->pgres);
         
@@ -776,15 +800,43 @@ _psyco_curs_execute(cursobject *self, char *query)
         self->description = PyTuple_New(pgnfields);
         self->casts = PyTuple_New(pgnfields);
         self->columns = pgnfields;
+
+        /* backend status message */
+        Py_XDECREF(self->status);
+        self->status = PyString_FromString(PQcmdStatus(self->pgres));
         
+		/* Calculate the display size for each column */
+#ifndef NO_DISPLAY_SIZE
+		dsize = (int *)calloc(pgnfields, sizeof(int));
+		if (dsize != NULL) {
+			if (self->rowcount == 0) {
+				for (i=0; i < pgnfields; i++)
+					dsize[i] = -1;
+			}
+			else {
+				int j, len;
+				for (j = 0; j < self->rowcount; j++) {
+					for (i = 0; i < pgnfields; i++) {
+                         len = PQgetlength(self->pgres, j, i);
+                         if (len > dsize[i]) dsize[i] = len;
+                    }
+				}
+			}
+		}
+#endif
+
         for (i = 0; i < pgnfields; i++) {
             /* int j, len = 0, maxl = 0; */
+            Oid ftype = PQftype(self->pgres, i);
+            int fsize = PQfsize(self->pgres, i);
+            int fmod =  PQfmod(self->pgres, i);
+            
             PyObject *dtitem = PyTuple_New(7);
-            PyObject *type = PyInt_FromLong(PQftype(self->pgres,i));
+            PyObject *type = PyInt_FromLong(ftype);
             PyObject *cast;
-            
+
             PyTuple_SET_ITEM(self->description, i, dtitem);
-            
+
             /* fill the right cast function by accessing the global
                dictionary of casting objects.  If we got no defined cast
                use the default one.
@@ -807,57 +859,81 @@ _psyco_curs_execute(cursobject *self, char *query)
                     cast, PQftype(self->pgres,i));
             Py_INCREF(cast);
             PyTuple_SET_ITEM(self->casts, i, cast);
-                
+
             /* fill the other fields */
             PyTuple_SET_ITEM(dtitem, 0,
-                             PyString_FromString(PQfname(self->pgres,i)));
+                             PyString_FromString(PQfname(self->pgres, i)));
             PyTuple_SET_ITEM(dtitem, 1, type);
-            
-            /* FIXME: display size is the maximum size of this field in
-               result tuples? what should we do with it? */
-            /* for (j = 0; j < self->rowcount; j++) {
-               len = PQgetlength(self->pgres, j, i);
-               maxl = ( maxl > len ? maxl : len);
-               }
-               PyTuple_SET_ITEM(dtitem, 2, PyInt_FromLong(maxl)); */
-            Py_INCREF(Py_None);
-            PyTuple_SET_ITEM(dtitem, 2, Py_None);
-            
+
+            /* display size is the maximum size of this field
+               result tuples. */
+            if (dsize && dsize[i] >= 0) {
+                PyTuple_SET_ITEM(dtitem, 2, PyInt_FromLong(dsize[i]));
+            }
+            else {
+                Py_INCREF(Py_None);
+                PyTuple_SET_ITEM(dtitem, 2, Py_None);
+            }
+
             /* size on the backend */
-            PyTuple_SET_ITEM(dtitem, 3,
-                             PyInt_FromLong(PQfsize(self->pgres, i)));
-            
-            /* FIXME: scale??? */
-            Py_INCREF(Py_None);
-            PyTuple_SET_ITEM(dtitem, 4, Py_None);
-            
-            /* FIXME: precision??? */
-            Py_INCREF(Py_None);
-            PyTuple_SET_ITEM(dtitem, 5, Py_None);
-            
+            if (fmod > 0) fmod = fmod - sizeof(int);
+            if (fsize == -1) {
+                if (ftype == NUMERICOID) {
+                    PyTuple_SET_ITEM(dtitem, 3,
+                                     PyInt_FromLong((fmod >> 16) & 0xFFFF));
+                }
+                else { /* If variable length record, return maximum size */
+                    PyTuple_SET_ITEM(dtitem, 3, PyInt_FromLong(fmod));
+                }
+            }
+            else {
+                PyTuple_SET_ITEM(dtitem, 3, PyInt_FromLong(fsize));
+            }
+
+            if (ftype == NUMERICOID) {
+                /* scale */
+                PyTuple_SET_ITEM(dtitem, 4,
+                                 PyInt_FromLong((fmod >> 16) & 0xFFFF));
+
+                /* precision */
+                PyTuple_SET_ITEM(dtitem, 5,
+                                 PyInt_FromLong(fmod & 0xFFFF));
+            }
+            else {
+                /* scale */
+                Py_INCREF(Py_None);
+                PyTuple_SET_ITEM(dtitem, 4, Py_None);
+
+                /* precision */
+                Py_INCREF(Py_None);
+                PyTuple_SET_ITEM(dtitem, 5, Py_None);
+            }
+
             /* FIXME: null_ok??? */
             Py_INCREF(Py_None);
             PyTuple_SET_ITEM(dtitem, 6, Py_None);
         }
+        
+		if (dsize) free(dsize);
     }
-        break;
+		break;
 
         /* ok but no tuples */
     case PGRES_COMMAND_OK:
         Dprintf("_psyco_curs_execute(): command returned OK (no tuples)\n");
-        /* FIXME: why we read a value that we _know_ being 0? */
+        /* sets rowcount from PQcmdTuples as an extra check */
         self->rowcount = atoi(PQcmdTuples(self->pgres));
         /* try to obtain the oid of an insert */
         self->last_oid = PQoidValue(self->pgres);
-        Py_XDECREF(self->description);
-        Py_INCREF(Py_None);
-        self->description = Py_None;
-        Py_XDECREF(self->casts);
-        self->casts = NULL;
-        self->notuples = 1;
-        break;
+
+        /* backend status message */
+        Py_XDECREF(self->status);
+        self->status = PyString_FromString(PQcmdStatus(self->pgres));
         
-            /* error! error! */
+        /* psyco_curs_reset already set the other fields */
+        break;
+    
+        /* error! error! */
     default: {
         char *pgerr = PQresultErrorMessage(self->pgres);
         char *errstr = NULL;
@@ -865,7 +941,7 @@ _psyco_curs_execute(cursobject *self, char *query)
         if (asprintf(&errstr, "%s\n%s", pgerr, query) >= 0) {
             if (!strncmp(pgerr, "ERROR:  Cannot insert a duplicate key", 37)
                 || !strncmp(pgerr, "ERROR:  ExecAppend: Fail to add null", 36)
-                || !strstr(pgerr, "referential integrity violation"))
+                || strstr(pgerr, "referential integrity violation"))
                 PyErr_SetString(IntegrityError, errstr);
             else
                 PyErr_SetString(ProgrammingError, errstr);
@@ -874,18 +950,26 @@ _psyco_curs_execute(cursobject *self, char *query)
         else {
             PyErr_SetString(ProgrammingError, pgerr);
         }
+			
+        Dprintf("_psyco_curs_execute(): error: errstr = %s\n",errstr);
+			
         CLEARPGRES(self->pgres);
         goto error;
     }
         break;
     }
+
+    /* check for result (should be at least None) */
+    if (res == NULL) {
+        Py_INCREF(Py_None);
+        res = Py_None;
+    }
     
     /* here we replace status with its default value */
     pthread_mutex_lock(&(self->keeper->lock));
     self->keeper->status = old_keeper_status;
     pthread_mutex_unlock(&(self->keeper->lock));
-    Py_INCREF(Py_None);
-    return Py_None;
+    return res;
     
   error:
     Dprintf("_psyco_curs_execute(): error, NOT resetting connection\n");
@@ -914,7 +998,8 @@ _mogrify_fmt(char *s, char x)
 
         case '(':
             while (*s && *s != ')') s++;
-            if (!*s || !*(++s)) return NULL;
+            while (*s && !isalpha(*s)) s++; 
+            if (!*s) return NULL;
             if (x) *s = x;
             return s;
             
@@ -974,7 +1059,6 @@ _mogrify_seq(PyObject *seq, PyObject *fmt)
         }
         else if (value == Py_None) {
             PyTuple_SET_ITEM(new_seq, i, PyString_FromString("NULL"));
-            Py_DECREF(Py_None);
             s = _mogrify_fmt(s, 's');
             Py_DECREF(value);
         }
@@ -1093,7 +1177,7 @@ psyco_curs_execute(cursobject *self, PyObject *args)
     Dprintf("psyco_curs_execute(): operation->refcnt = %d\n",
             operation->ob_refcnt);
 
-    res = _psyco_curs_execute(self, query);
+    res = _psyco_curs_execute(self, query, NULL, NULL);
     free(query);
     return res;
 }
@@ -1193,7 +1277,7 @@ psyco_curs_callproc(cursobject *self, PyObject *args)
     }
 
 
-    _psyco_curs_execute(self, query);
+    _psyco_curs_execute(self, query, NULL, NULL);
     free(query);
     return seq;
 }
@@ -1209,7 +1293,7 @@ static PyObject *
 psyco_curs_executemany(cursobject *self, PyObject *args)
 {
     PyObject *tmpobj, *parm_seq = NULL, *seq_item = NULL;
-    PyObject *operation = NULL;
+    PyObject *operation = NULL, *last = NULL;
     int i;
 
     if (!PyArg_ParseTuple(args, "O!O&",
@@ -1221,26 +1305,33 @@ psyco_curs_executemany(cursobject *self, PyObject *args)
     EXC_IFCLOSED(self);
     
     tmpobj = PyTuple_New(2);
-
+    
     Py_INCREF(operation);
     PyTuple_SET_ITEM(tmpobj, 0, operation);
-
     for (i = 0; i < PyTuple_Size(parm_seq); i++) {
         seq_item = PySequence_GetItem(parm_seq, i); 
         if (!PyDict_Check(seq_item) && !PyTuple_Check(seq_item)) {
             PyErr_SetString(PyExc_TypeError,
                             "arg 2 must be a dictionary or tuple sequence");
+	    Py_DECREF(tmpobj);
+	    Py_DECREF(seq_item);
+	    Py_DECREF(parm_seq);
             return NULL;
         }
-        Py_INCREF(seq_item);
         PyTuple_SET_ITEM(tmpobj, 1, seq_item);
+	if (last != NULL) {
+        Py_DECREF(last);
+	}
+	last = seq_item;
         if(!psyco_curs_execute(self, tmpobj)) {
             Py_DECREF(tmpobj);
+            Py_DECREF(parm_seq);
             return NULL;
         }
     }
 
     Py_DECREF(tmpobj);
+    Py_DECREF(parm_seq);
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -1295,8 +1386,10 @@ psyco_curs_fetchone(cursobject *self, PyObject *args)
         Dprintf("psyco_curs_fetchone: str->refcnt = %d\n", str->ob_refcnt);
         PyTuple_SET_ITEM(arg, 0, str);
         val = PyObject_CallObject(PyTuple_GET_ITEM(self->casts, i), arg);    
-        Dprintf("psyco_curs_fetchone: val->refcnt = %d\n", val->ob_refcnt);
-        Dprintf("psyco_curs_fetchone: str->refcnt = %d\n", str->ob_refcnt);
+        if (val) {
+            Dprintf("psyco_curs_fetchone: val->refcnt = %d\n", val->ob_refcnt);
+        }
+            Dprintf("psyco_curs_fetchone: str->refcnt = %d\n", str->ob_refcnt);
         Py_DECREF(arg);
         
         if (val) {
@@ -1413,7 +1506,7 @@ psyco_curs_fetchall(cursobject *self, PyObject *args)
 }
 
 
-/* psyco_curs_dictfetchone() - fetch onw row of data in a dictionary */
+/* psyco_curs_dictfetchone() - fetch one row of data in a dictionary */
 
 static char psyco_curs_dictfetchone__doc__[] = 
 "Fetch the next row of a query result set, returning a "
@@ -1517,7 +1610,6 @@ psyco_curs_dictfetchall(cursobject *self, PyObject *args)
     PyObject *res, *list;
     int size, i;
 
-    PARSEARGS(args);
     EXC_IFCLOSED(self);
     EXC_IFNOTUPLES(self);
     
@@ -1558,7 +1650,6 @@ static char psyco_curs_setinputsizes__doc__[] =
 static PyObject *
 psyco_curs_setinputsizes(cursobject *self, PyObject *args)
 {
-    PARSEARGS(args);
     EXC_IFCLOSED(self);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1569,12 +1660,13 @@ psyco_curs_setinputsizes(cursobject *self, PyObject *args)
 
 static char psyco_curs_setoutputsizes__doc__[] = 
 "Set a column buffer size for fetches of large columns.\n"
-"The column is specified as an index into the result sequence. Not specifying\n"
-"the column will set the default size for all large columns in the cursor.\n"
+"The column is specified as an index into the result sequence. Not\n"
+"specifying the column will set the default size for all large columns\n"
+"in the cursor.\n"
 "At now this method does nothing.";
 
 static PyObject *
-psyco_curs_setoutputsizes(cursobject *self, PyObject *args)
+psyco_curs_setoutputsize(cursobject *self, PyObject *args)
 {
     PARSEARGS(args);
     EXC_IFCLOSED(self);
@@ -1618,9 +1710,12 @@ psyco_curs_autocommit(cursobject *self, PyObject *args)
 
     Dprintf("psyco_curs_autocommit(): autocommit = %ld on %p\n", ac, self);
 
+    if (ac) ac = 0;
+    else ac = 2;
+
     EXC_IFNOTONLY(self);
     EXC_IFCRITICAL(self);
-    curs_switch_autocommit(self, ac);
+    curs_switch_isolation_level(self, ac);
     EXC_IFCRITICAL(self);
 
     Py_INCREF(Py_None);
@@ -1652,6 +1747,131 @@ psyco_curs_lastoid(cursobject *self, PyObject *args)
 }
 
 
+/* psyco_curs_copy_to() - copy a table to a local file */
+
+static char psyco_curs_copy_to__doc__[] =
+"Copy table to a local file.\ncopy_to(fileObject, tablename, <separator>, <null identifier>)""";
+
+static PyObject *
+_psyco_curs_copy_to(cursobject *self, PyObject *file)
+{
+    char buffer[4096];
+    int status, len;
+    PyObject *o;
+    
+    while (1) {
+        status = PQgetline(self->pgconn, buffer, 4096);
+        if (status == 0) {
+            if (buffer[0] == '\\' && buffer[1] == '.') break;
+            
+            len = strlen(buffer);
+            buffer[len++] = '\n';
+        }
+        else if (status == 1) {
+            len = 4096-1;
+        }
+        else {
+            return NULL;
+        }
+
+        o = PyString_FromStringAndSize(buffer, len);
+        PyObject_CallMethod(file, "write", "O", o);
+        Py_DECREF(o);
+    }
+
+    PQendcopy(self->pgconn);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+     
+static PyObject *
+psyco_curs_copy_to(cursobject *self, PyObject *args)
+{
+    char *table_name, *query = NULL;
+    char *sep = "\t", *null =NULL;
+    PyObject *file;
+
+    if (!PyArg_ParseTuple(args, "O!s|ss", 
+                          &PyFile_Type, &file, &table_name,
+                          &sep, &null)) {
+        return NULL;
+    }
+    EXC_IFCRITICAL(self);
+
+    if (null) {
+        asprintf(&query, "COPY %s TO stdout USING DELIMITERS '%s'"
+                     " WITH NULL AS '%s'", table_name, sep, null);
+    }
+    else {
+        asprintf(&query, "COPY %s TO stdout USING DELIMITERS '%s'" ,
+                 table_name, sep);
+    }
+    Dprintf("psyco_curs_copy_to(): query = %s\n", query);
+    
+    _psyco_curs_execute(self, query, _psyco_curs_copy_to, file);
+    free(query);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+/* psyco_curs_copy_from() - copy a table from a local file */
+
+static char psyco_curs_copy_from__doc__[] =
+"Copy table from a local file.\ncopy_from(fileObject, tablename, <separator>, <null identifier>)""";
+
+static PyObject *
+_psyco_curs_copy_from(cursobject *self, PyObject *file)
+{
+    PyObject *o;
+  	
+    while (1) {
+        o = PyObject_CallMethod(file, "readline", NULL);
+        if (!o || o == Py_None || PyString_GET_SIZE(o) == 0) break;
+        if (PQputline(self->pgconn, PyString_AS_STRING(o)) != 0) {
+            Py_DECREF(o);
+            return NULL;
+        }
+        Py_DECREF(o);
+    }
+    PQputline(self->pgconn, "\\.\n");
+    PQendcopy(self->pgconn);
+   
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+     
+static PyObject *
+psyco_curs_copy_from(cursobject *self, PyObject *args)
+{
+    char *table_name, *query = NULL;
+    char *sep = "\t", *null = NULL;
+    PyObject *file;
+
+    if (!PyArg_ParseTuple(args, "O!s|ss", 
+                          &PyFile_Type, &file, &table_name,
+                          &sep, &null)) {
+        return NULL;
+    }
+    EXC_IFCRITICAL(self);
+
+    if (null) {
+        asprintf(&query, "COPY %s FROM stdin USING DELIMITERS '%s'"
+                     " WITH NULL AS '%s'", table_name, sep, null);
+    }
+    else {
+        asprintf(&query, "COPY %s FROM stdin USING DELIMITERS '%s'" , table_name, sep);
+    }
+    Dprintf("psyco_curs_copy_from(): query = %s\n", query);
+    
+    _psyco_curs_execute(self, query, _psyco_curs_copy_from, file);
+    free(query);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
 /**** CURSOR OBJECT DEFINITION ****/
 
 /* object methods list */
@@ -1683,7 +1903,7 @@ static struct PyMethodDef psyco_curs_methods[] = {
      METH_VARARGS, psyco_curs_dictfetchall__doc__},
     {"setinputsizes", (PyCFunction)psyco_curs_setinputsizes,
      METH_VARARGS, psyco_curs_setinputsizes__doc__},
-    {"setoutputsizes", (PyCFunction)psyco_curs_setoutputsizes,
+    {"setoutputsize", (PyCFunction)psyco_curs_setoutputsize,
      METH_VARARGS, psyco_curs_setoutputsizes__doc__},
     {"nextset", (PyCFunction)psyco_curs_nextset,
      METH_VARARGS, psyco_curs_nextset__doc__},
@@ -1691,6 +1911,10 @@ static struct PyMethodDef psyco_curs_methods[] = {
      METH_VARARGS, psyco_curs_autocommit__doc__},
     {"lastoid", (PyCFunction)psyco_curs_lastoid,
      METH_VARARGS, psyco_curs_lastoid__doc__},
+    {"copy_to", (PyCFunction)psyco_curs_copy_to,
+     METH_VARARGS, psyco_curs_copy_to__doc__},
+    {"copy_from", (PyCFunction)psyco_curs_copy_from,
+     METH_VARARGS, psyco_curs_copy_from__doc__},
     {NULL, NULL}
 };
 
@@ -1703,6 +1927,8 @@ static struct memberlist psyco_curs_memberlist[] = {
     {"rowcount", T_LONG, OFFSETOF(rowcount), RO},
     {"arraysize", T_LONG, OFFSETOF(arraysize), 0},
     {"description", T_OBJECT,OFFSETOF(description), RO},
+    {"lastrowid", T_LONG,OFFSETOF(last_oid), RO},
+    {"statusmessage", T_OBJECT,OFFSETOF(status), RO},
     {NULL}	
 };
 
@@ -1740,7 +1966,8 @@ psyco_curs_destroy(cursobject *self)
     /* destroy cursor information */
     _psyco_curs_destroy(self);
     Py_XDECREF(self->description);
-
+    Py_XDECREF(self->status);
+    
     Dprintf("psyco_curs_destroy(): cursor at %p destroyed, refcnt = %d\n",
             self, self->ob_refcnt);
     PyObject_Del(self);
@@ -1752,7 +1979,7 @@ static char Cusrtype__doc__[] =
 "used to manage the context of a fetch operation.";
 
 static PyTypeObject Curstype = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -1794,7 +2021,7 @@ new_psyco_cursobject(connobject *conn, connkeeper *keeper)
 {
     cursobject *self;
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     Curstype.ob_type = &PyType_Type;
 #endif
@@ -1811,12 +2038,14 @@ new_psyco_cursobject(connobject *conn, connkeeper *keeper)
     self->rowcount = -1;
     self->closed = 0;
     self->last_oid = InvalidOid;
-    self->autocommit = conn->autocommit ? 1 : 0;
+    self->isolation_level = conn->isolation_level;
     self->casts = NULL;
     self->notice = NULL;
     self->critical = NULL;
     self->description = Py_None;
     Py_INCREF(Py_None);
+    self->status = Py_None;
+    Py_INCREF(Py_None);
     
     /* get a connection to db */
     if (keeper != NULL) {   
