@@ -21,7 +21,7 @@
 
 /*
  * module.h -- defines constants needed by the module interface DBAPI 2.0
- * $Id: module.h,v 1.48.2.1 2002/04/21 22:58:52 fog Exp $
+ * $Id: module.h,v 1.52 2003/05/07 13:27:17 fog Exp $
 */
 
 #ifndef __PSYCOPG_MODULE__
@@ -60,6 +60,11 @@ static int pthread_mutex_init(pthread_mutex_t *mutex, void* temp)
 #define PyObject_Del PyMem_DEL
 #endif
 
+/**** PyObject_TypeCheck introduced in 2.2 ****/
+#ifndef PyObject_TypeCheck
+#define PyObject_TypeCheck(o, t) ((o)->ob_type == (t))
+#endif
+
 /**** globals of the module ****/
 #define APILEVEL "2.0"
 #define THREADSAFETY 2
@@ -82,12 +87,16 @@ static int pthread_mutex_init(pthread_mutex_t *mutex, void* temp)
 
 
 /**** debug printf-like function ****/
+#if defined( __GNUC__) && !defined(__APPLE__)
 #ifdef PSYCOTIC_DEBUG
-#ifdef __GNUC__
 #include <sys/types.h>
 #include <unistd.h>
 #define Dprintf(fmt, args...) fprintf(stderr, "[%d] " fmt, getpid() , ## args)
-#else /* !__GNUC__ */
+#else
+#define Dprintf(fmt, args...)
+#endif 
+#else /* !__GNUC__ or __APPLE__ */
+#ifdef PSYCOTIC_DEBUG
 #include <stdarg.h>
 static void Dprintf(const char *fmt, ...)
 {
@@ -96,10 +105,6 @@ static void Dprintf(const char *fmt, ...)
     vprintf(fmt, ap);
     va_end(ap);
 }
-#endif
-#else
-#ifdef __GNUC__
-#define Dprintf(fmt, args...)
 #else
 static void Dprintf(const char *fmt, ...) {}
 #endif
@@ -138,7 +143,7 @@ typedef struct {
     int closed;           /* 1 if the connection is closed */
     int maxconn;          /* max connections to db */
     int minconn;          /* minimum number of open connections */
-    int autocommit;       /* default value for autocommit */
+    int isolation_level;  /* isolation level */
     int serialize;        /* serialize cursors? */
 } connobject;
 
@@ -179,11 +184,14 @@ struct _cursobject {
     /* an array (tuple) of typecast functions */
     PyObject *casts;
 
+    /* last message from the server after an execute */
+    PyObject *status;
+    
     /* last oid from an insert or InvalidOid */
     Oid last_oid;
 
-    /* autocommit, make cursor do an implicit commit after each transaction */
-    int autocommit;
+    /* isolation level for this cursor */
+    int isolation_level;
 
     /* a notice from the backend */
     char *notice;
@@ -194,18 +202,22 @@ struct _cursobject {
 
 cursobject *new_psyco_cursobject(connobject *conn, connkeeper *keeper);
 
+/**** function type used in execute callbacks ****/
+typedef PyObject *(*_psyco_curs_execute_callback)(cursobject *s, PyObject *o);
+
 
 /**** mxDateTime API access ****/
 extern mxDateTimeModule_APIObject *mxDateTimeP;
 
 
 /**** transaction control functions ****/
+extern int dispose_pgconn(cursobject *self);
 extern int commit_pgconn(cursobject *self);
 extern int abort_pgconn(cursobject *self);
 extern int begin_pgconn(cursobject *self);
 
 /**** other globally visible functions ****/
-extern void curs_switch_autocommit(cursobject *self, long int ac);
+extern void curs_switch_isolation_level(cursobject *self, long int level);
 extern connkeeper *alloc_keeper(connobject *conn);
 
 
