@@ -20,12 +20,13 @@
 
 /*
  * typemod.c -- types available to the psycopg module
- * $Id: typemod.c,v 1.9.2.3 2002/04/23 17:37:21 fog Exp $
+ * $Id: typemod.c,v 1.13 2003/05/07 13:27:17 fog Exp $
  */
 
 #ifndef _WIN32
 #include <unistd.h>
 #endif
+
 #include <assert.h>
 
 #include "module.h"
@@ -92,7 +93,7 @@ static char psyco_DateTimeObject__doc__[] =
 "Object to pass (in a portable way) dates and times to the database.";
 
 PyTypeObject psyco_DateTimeObject_Type = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -126,7 +127,7 @@ new_psyco_datetimeobject(PyObject *datetime, int type)
 {
     psyco_DateTimeObject *obj;
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     psyco_DateTimeObject_Type.ob_type = &PyType_Type;
 #endif
@@ -134,7 +135,6 @@ new_psyco_datetimeobject(PyObject *datetime, int type)
     obj = PyObject_NEW(psyco_DateTimeObject, &psyco_DateTimeObject_Type);
     if (obj == NULL) return NULL;
 
-    Py_INCREF(datetime);
     obj->datetime = datetime;
     obj->type = type;
 
@@ -198,7 +198,7 @@ static char psyco_BufferObject__doc__[] =
 "Object to pass (in a portable way) binary data to the database.";
 
 PyTypeObject psyco_BufferObject_Type = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -234,7 +234,7 @@ new_psyco_bufferobject(PyObject *buffer)
     unsigned char *original, *quoted, *chptr, *newptr;
     int i, len, space, new_space;
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     psyco_BufferObject_Type.ob_type = &PyType_Type;
 #endif
@@ -377,7 +377,7 @@ static char psyco_QuotedStringObject__doc__[] =
 "A string that quotes itself following PostgreSQL quoting convention.";
 
 PyTypeObject psyco_QuotedStringObject_Type = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -413,7 +413,7 @@ new_psyco_quotedstringobject(PyObject *buffer)
     char *original, *quoted;
     int i, j, len;
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     psyco_QuotedStringObject_Type.ob_type = &PyType_Type;
 #endif
@@ -599,6 +599,7 @@ psyco_DateFromMx(PyObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "O!", mxDateTimeP->DateTime_Type, &o))
         return NULL;
+    Py_INCREF(o);
     return new_psyco_datetimeobject(o, PSYCO_DATETIME_DATE);
 }
 
@@ -609,6 +610,7 @@ psyco_TimeFromMx(PyObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "O!", mxDateTimeP->DateTime_Type, &o))
         return NULL;
+    Py_INCREF(o);
     return new_psyco_datetimeobject(o, PSYCO_DATETIME_TIME);
 }
 
@@ -619,6 +621,7 @@ psyco_TimestampFromMx(PyObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "O!", mxDateTimeP->DateTime_Type, &o))
         return NULL;
+    Py_INCREF(o);
     return new_psyco_datetimeobject(o, PSYCO_DATETIME_TIMESTAMP);
 }
 
