@@ -20,13 +20,16 @@
 
 /*
  * typeobj.c -- define the type objects
- * $Id: typeobj.c,v 1.20.2.14 2002/10/25 10:12:56 fog Exp $
+ * $Id: typeobj.c,v 1.24 2003/03/06 10:13:15 fog Exp $
  */
 
 #include "assert.h"
 #include "module.h"
 #include "typeobj.h"
 
+static PyObject *psyco_DBAPITypeObject_new(PyObject *, PyObject *, PyObject *);
+extern PyTypeObject psyco_DBAPITypeObject_Type;
+
 
 /**** builtin C casting functions ****/
 
@@ -132,7 +135,7 @@ skip_until_space(char *s)
 static PyObject *
 psyco_INTERVAL_cast(PyObject *s)
 {
-    long years = 0, months = 0, days = 0;
+    long years = 0, months = 0, days = 0, denominator = 1;
     double hours = 0.0, minutes = 0.0, seconds = 0.0, hundredths = 0.0;
     double v = 0.0, sign = 1.0;
     int part = 0;
@@ -154,6 +157,11 @@ psyco_INTERVAL_cast(PyObject *s)
         case '5': case '6': case '7': case '8': case '9':
             v = v*10 + (double)*str - (double)'0';
             Dprintf("psyco_INTERVAL_cast(): v = %f\n", v);
+            if (part == 6){
+                denominator *= 10;
+                Dprintf("psyco_INTERVAL_cast(): denominator = %ld\n",
+                        denominator);
+            }
             break;
 
         case 'y':
@@ -225,17 +233,17 @@ psyco_INTERVAL_cast(PyObject *s)
     }
     else if (part == 6) {
         hundredths = v;
-        Dprintf("psyco_INTERVAL_cast(): hundredths = %f\n",
-                hundredths);
+        Dprintf("psyco_INTERVAL_cast(): hundredths = %f\n", hundredths);
+        hundredths = hundredths/denominator;
+        Dprintf("psyco_INTERVAL_cast(): ACTUAL FRACTIONS = %.20f\n", hundredths);
     }
     
-
     /* calculates seconds */
     if (hours < 0.0) {
-        seconds = - (hundredths/100.0 + seconds + minutes*60) + hours*3600;
+        seconds = - (hundredths + seconds + minutes*60) + hours*3600;
     }
     else {
-        seconds += hundredths/100.0 + minutes*60 + hours*3600;
+        seconds += hundredths + minutes*60 + hours*3600;
     }
 
     /* calculates days */
@@ -400,26 +408,27 @@ static struct memberlist psyco_DBAPITypeObject_memberlist[] = {
 
 static int
 psyco_DBAPITypeObject_coerce(PyObject **pv, PyObject **pw)
-{
-    static PyObject *psyco_DBAPITypeObject_new(PyObject *, PyObject *,
-                                               PyObject *);
-
-    if (PyInt_Check(*pw)) {
-        PyObject *coer, *args;
-        args = PyTuple_New(1);
-        Py_INCREF(*pw);
-        PyTuple_SET_ITEM(args, 0, *pw);
-        coer = psyco_DBAPITypeObject_new(NULL, args, NULL);
-        *pw = coer;
-        Py_DECREF(args);
-        Py_INCREF(*pv);
-        return 0;
-    }
-    else {
-        Py_INCREF(*pv);
-        Py_INCREF(*pw);
-        return 0;
+{    
+    if (PyObject_TypeCheck(*pv, &psyco_DBAPITypeObject_Type)) {
+        if (PyInt_Check(*pw)) {
+            PyObject *coer, *args;
+            args = PyTuple_New(1);
+            Py_INCREF(*pw);
+            PyTuple_SET_ITEM(args, 0, *pw);
+            coer = psyco_DBAPITypeObject_new(NULL, args, NULL);
+            *pw = coer;
+            Py_DECREF(args);
+            Py_INCREF(*pv);
+            return 0;
+        }
+        else if (PyObject_TypeCheck(*pw, &psyco_DBAPITypeObject_Type)){
+            Py_INCREF(*pv);
+            Py_INCREF(*pw);
+            return 0;
+        }
     }
+    PyErr_SetString(PyExc_TypeError, "psycopg type coercion failed");
+    return -1;
 }
 
 static PyNumberMethods psyco_DBAPITypeObject_as_number = {
@@ -556,7 +565,7 @@ psyco_DBAPITypeObject_call(PyObject *self, PyObject *args, PyObject *keywds)
 static char psyco_DBAPITypeObject__doc__[] = "Object to classify data types.";
 
 PyTypeObject psyco_DBAPITypeObject_Type = {
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__CYGWIN__)
     PyObject_HEAD_INIT(&PyType_Type)
 #else
     PyObject_HEAD_INIT(NULL)
@@ -590,7 +599,7 @@ psyco_DBAPITypeObject_new(PyObject *name, PyObject *values, PyObject *cast)
 {
     psyco_DBAPITypeObject *obj;
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     /* For MSVC workaround */
     psyco_DBAPITypeObject_Type.ob_type = &PyType_Type;
 #endif
